#!/bin/sh
append DRIVERS "rt3052"

prepare_config() {
	local device=$1
	config_get channel $device channel
	config_get ssid1 private ssid
	config_get ssid2 public ssid
	config_get mode $device mode

	mkdir -p /var/etc
	cat > /var/etc/RT2860.dat<<EOF
#The word of "Default" must not be removed
Default
CountryRegion=0
CountryRegionABand=0
CountryCode=US
BssidNum=1
SSID1=${ssid1:-MyPlace}
SSID2=${ssid2:-FON_FREE_INTERNET}
WirelessMode=${mode:-9}
TxRate=0
Channel=${channel:-8}
BasicRate=15
BeaconPeriod=100
DtimPeriod=1
TxPower=90
DisableOLBC=0
BGProtection=0
TxAntenna=
RxAntenna=
TxPreamble=0
RTSThreshold=2347
FragThreshold=2346
TxBurst=1
PktAggregate=0
TurboRate=0
WmmCapable=0
APAifsn=3;7;1;1
APCwmin=4;4;3;2
APCwmax=6;10;4;3
APTxop=0;0;94;47
APACM=0;0;0;0
BSSAifsn=3;7;2;2
BSSCwmin=4;4;3;2
BSSCwmax=10;10;4;3
BSSTxop=0;0;94;47
BSSACM=0;0;0;0
AckPolicy=0;0;0;0
NoForwarding=0
NoForwardingBTNBSSID=0
HideSSID=0
ShortSlot=1
AutoChannelSelect=0
IEEE8021X=0
IEEE80211H=0
CSPeriod=10
WirelessEvent=0
PreAuth=0
AuthMode=OPEN
EncrypType=NONE
RekeyInterval=0
RekeyMethod=DISABLE
PMKCachePeriod=10
WPAPSK=
DefaultKeyID=1
Key1Type=0
Key1Str=
Key2Type=0
Key2Str=
Key3Type=0
Key3Str=
Key4Type=0
Key4Str=
HSCounter=0
AccessPolicy0=0
AccessControlList0=
AccessPolicy1=0
AccessControlList1=
AccessPolicy2=0
AccessControlList2=
AccessPolicy3=0
AccessControlList3=
HT_HTC=0
HT_RDG=1
HT_EXTCHA=0
HT_LinkAdapt=0
HT_OpMode=0
HT_MpduDensity=5
HT_BW=1
HT_AutoBA=1
HT_AMSDU=0
HT_BAWinSize=64
HT_GI=1
HT_STBC=1
HT_MCS=33
EOF
}

scan_rt3052() {
	local device="$1"
}

disable_rt3052() {
	local device="$1"
	config_get vifs "$device" vifs
	for vif in $vifs; do
		config_get ifname $vif ifname
		ifconfig $ifname down
	done
}

enable_rt3052() {
	local device="$1"
	prepare_config $device
	config_get vifs "$device" vifs
	for vif in $vifs; do
		ifconfig $ifname down
	done		
	config_get disabled "$device" disabled
	[ "$disabled" = "1" ] && return
	for vif in $vifs; do
		local ifname encryption key ssid
		config_get ifname $vif ifname
		config_get encryption $vif encryption
		config_get key $vif key
		[ "$key" = "" -a "$vif" = "private" ] && {
			logger "no key set serial"
			key="AAAAAAAAAA"
		}
		config_get ssid $vif ssid
		config_get wpa_crypto $vif wpa_crypto
		echo "......." $ssid
		[ "$ssid" == "off" ] && echo ifconfig $ifname down
		[ "$ssid" == "off" ] || 
		{
			ifconfig $ifname up
			case "$encryption" in
				wpa*|WPA*|Mixed|mixed)
					local enc
					case "$encryption" in
						WPA|WPA1|wpa|wpa1) enc=WPAPSK;;
						WPA2|wpa2) enc=WPA2PSK;;
						Mixed|mixed) enc=WPAPSKWPA2PSK;;
					esac
					local crypto="TKIP"
					case "$wpa_crypto" in
						aes|AES) crypto=AES;;
						tkip+aes|TKIP+AES) crypto=TKIPAES;;
					esac
					iwpriv $ifname set AuthMode=$enc
					iwpriv $ifname set EncrypType=$crypto
					iwpriv $ifname set IEEE8021X=0
					iwpriv $ifname set SSID=${ssid}
					iwpriv $ifname set WPAPSK=${key}
					iwpriv $ifname set DefaultKeyID=2
					iwpriv $ifname set SSID=${ssid}
					;;
				WEP|wep)
					iwpriv $ifname set AuthMode=SHARED
					iwpriv $ifname set EncrypType=WEP
					iwpriv $ifname set IEEE8021X=0
					iwpriv $ifname set Key1=${key}
					iwpriv $ifname set DefaultKeyID=1
					iwpriv $ifname set SSID=${ssid}
					;;
				none|open)
					;;
			esac
			local net_cfg bridge
			net_cfg="$(find_net_config "$vif")"
			[ -z "$net_cfg" ] || {
				bridge="$(bridge_interface "$net_cfg")"
				config_set "$vif" bridge "$bridge"
				start_net "$ifname" "$net_cfg"
			}
			set_wifi_up "$vif" "$ifname"
		}
	done
}

detect_rt3052() {
	cd /sys/module/
	[ -d rt2860v2_ap ] || return
	cat <<EOF
config wifi-device	rt3052
	option type     rt3052
	option channel  auto
	option disabled 0

config wifi-iface	private
	option device   rt3052
	option ifname	ra0
	option network  lan
	option ssid 	OpenWrt
	option encryption wpa2
	option key 	0123456789

#config wifi-iface	public
#	option device   rt3052
#	option ifname	ra1
#	option network  lan
#	option ssid 	off
#	option encryption none
EOF
}


